﻿using Dam.Controllers;
using Dam.Core;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;

namespace Dam
{
    public class BandaidGraphic
    {
        public Bandaid Bandaid { get; set; }
        public Texture2D Graphic { get; }
        public bool Visible { get; set; }

        public BandaidGraphic(GraphicsDevice gd, Bandaid p)
        {
            Bandaid = p;
            Graphic = DamGraphicsController.Bandaid.GetRandomTexture();
            Visible = false;
        }
    }

    public class Bandaid
    {
        public int X => (int)realX;
        public int Y => (int)realY;

        public int StuckX { get; set; }

        public const int WIDTH = 49;
        public const int HEIGHT = 95;
        public const float FLOOR_FRICTION = 0.7f;
        public const float BOUNCE = 0.5f;

        private float gravity = 1.5f;
        public float realX;
        public float realY;
        public float velocityX;
        public float velocityY;

        public float rotateSpeed;
        public float rotation;

        public bool isOnWall;
        public bool isAlmostOnWall; // Final approach animation is playing.
        public bool isAlmostOnWallFirstFrame; // Final approach animation is playing.

        public TimeSpan TimeAlive { get; set; }

        public bool markForDeletion;
        public bool markForNotBounce;

        public float VelocityMag
        {
            get
            {
                return (float)Math.Sqrt(Math.Pow(velocityX, 2) + Math.Pow(velocityY, 2));
            }
        }

        public float VelocityDir
        {
            get
            {
                return (float)Math.Atan2(velocityY, velocityX);
            }
        }

        public Vector2 Velocity
        {
            get
            {
                var angle = Math.Atan2(velocityY, velocityX);
                return new Vector2((float)Math.Cos(angle), (float)Math.Sin(angle));
            }
        }

        public Bandaid(int x, int y, bool isLeft)
        {
            Reset(x, y, isLeft);
        }

        public void Reset(int x, int y, bool isLeft)
        {
            realX = x;
            realY = y;
            velocityX = DamGameController.random.Next(14, 18) / 1f * (isLeft ? -1 : 1);
            velocityY = -20f;

            // rotate, random speed random direction.
            rotateSpeed = DamGameController.random.Next(4, 7) * (DamGameController.random.Next(0, 2) == 0 ? -1 : 1) / 50f;

            TimeAlive = TimeSpan.Zero;

            framesSinceLargeVelocity = 0;
            smallVelocity = 5f;

            isOnWall = false;
            isAlmostOnWall = false;
            isAlmostOnWallFirstFrame = false;

            markForDeletion = false;
            markForNotBounce = false;
        }

        int framesSinceLargeVelocity;
        float smallVelocity;

        public void Update(GameTime gt, DamGameController gc, float cameraOffset)
        {
            if(isAlmostOnWall && isAlmostOnWallFirstFrame)
            {
                isAlmostOnWallFirstFrame = false;
            }

            TimeAlive += gt.ElapsedGameTime;

            var currentY = Y;
            var newY = (int)(realY + velocityY);
            
            if(!(isAlmostOnWall || isOnWall))
            {
                velocityY += gravity;

                // These checks don't use half the width, because we want bandaids to bounce on their centre (same for floor)// Left wall check.
                if (realX < cameraOffset)
                {
                    realX = cameraOffset;
                    velocityX *= -FLOOR_FRICTION;
                }
                // right wall check
                else if (realX > DamGame.BASE_GAME_RES_WIDTH + cameraOffset)
                {
                    realX = DamGame.BASE_GAME_RES_WIDTH + cameraOffset;
                    velocityX *= -FLOOR_FRICTION;
                }
                // MinX is the closest segment. Don't need to check all segments until this is true.
                else if (realX + WIDTH / 2 > gc.Wall.MinX)
                {
                    Console.WriteLine("Potential wall hit.....");

                    // We might have hit the wall... check each segment.

                    for (int i = 0; i < gc.Wall.Segments.Length; i++)
                    {
                        var segment = gc.Wall.Segments[i];
                        var yPos = WallSegment.GetSegmentY(i);

                        if (realX + WIDTH / 2 > segment.X && realX - WIDTH / 2 < segment.X && yPos + Wall.SEGMENT_HEIGHT / 2 > Y - HEIGHT / 2 && yPos - Wall.SEGMENT_HEIGHT / 2 < Y + HEIGHT / 2)
                        //if (realX + 10 > segment.X)
                        {
                            // We hit a segment.
                            // Bounce off
                            //realX = segment.X;
                            //velocityX *= -FLOOR_FRICTION;

                            foreach (var c in gc.Wall.Cracks)
                            {
                                if (c.IsActive)
                                {
                                    if (!c.IsCovered && c.Y > Y - HEIGHT / 2 && c.Y < Y + HEIGHT / 2)
                                    {
                                        c.IsCovered = true;

                                        // Only let bandaid stick if it covers a crack.
                                        // We hit the dam wall.
                                        isAlmostOnWall = true;
                                        isAlmostOnWallFirstFrame = true;

                                        StuckX = segment.X;
                                    }
                                }
                            }

                            if (isAlmostOnWall)
                            {
                                // Stop checking segments. FIXME don't want a return here?
                                return;
                            }
                        }
                    }
                }

                // Floor check
                if (realY >= DamGame.BASE_GAME_RES_HEIGHT - DamGameController.FLOOR_DISTANCE_FROM_BOTTOM_OF_SCREEN)
                {
                    if (velocityY > 0.5f && !markForNotBounce)
                    {
                        DamSoundController.Play(DamSoundController.LowBoop);
                    }

                    realY = DamGame.BASE_GAME_RES_HEIGHT - DamGameController.FLOOR_DISTANCE_FROM_BOTTOM_OF_SCREEN;
                    velocityX *= FLOOR_FRICTION;
                    velocityY *= markForNotBounce ? 0 : -BOUNCE;
                }

                if (Math.Abs(velocityX) < 0.5f)
                {
                    velocityX = 0;
                }
            }
            
            if (markForNotBounce)
            {
                rotateSpeed *= 0.85f;
            }

            rotation += rotateSpeed;

            if (Math.Abs(velocityY) > smallVelocity)
            {
                framesSinceLargeVelocity = 0;
            }
            framesSinceLargeVelocity++;
            if (framesSinceLargeVelocity > 10)
            {
                markForNotBounce = true;
            }
            if (framesSinceLargeVelocity > 60)
            {
                markForDeletion = true;
            }

            if (isAlmostOnWall)
            {
                velocityX *= 0.3f;
                velocityY *= 0.3f;
                rotateSpeed *= 0.3f;

                if(Math.Abs(velocityX) < 0.2f)
                {
                    isOnWall = true;
                }
            }

            realX += velocityX;
            realY += velocityY;
        }
    }
}
