﻿using Microsoft.Xna.Framework;
using Spl.Core;
using Spl.Core.Input;
using System;
using System.Linq;
using IC = Dam.Controllers.DamInputController;

namespace Dam.Controllers
{
    public class DamGameController : GameController
    {
        private CameraController _cameraController;

        public static GameState MainMenu;
        public static GameState LevelInitialising;
        public static GameState LevelRunning;
        public static GameState GameOver;

        public enum MenuOption { NewGame, ToggleMute, Exit }
        public MenuOption currentMainMenuOption = MenuOption.NewGame;

        public const int FLOOR_DISTANCE_FROM_BOTTOM_OF_SCREEN = 70;
        public const float BANDAID_COOLDOWN_IN_SECONDS = 0.5f;
        public const float BANDAID_MAX_LIFETIME_IN_SECONDS = 5f;
        public const int MAX_WATER_PARTICLES = 5000; // Appears to be roughly the max shown particles.
        public const int MAX_BANDAIDS = (int)(BANDAID_MAX_LIFETIME_IN_SECONDS / BANDAID_COOLDOWN_IN_SECONDS);
        public const int NUM_FRAMES_MOVEMENT_DISABLED = 5;

        // Game objects.
        public Wall Wall { get; set; }
        public Hero Hero { get; set; }
        public BandaidGraphic[] Bandaids { get; set; }
        public WaterGraphic[] Waters { get; set; }

        public BackgroundLayer[] BackgroundLayers { get; set; }

        // Timers. Timer = time until X. Time = A set time to count up to.
        public TimeSpan CrackTimer { get; set; }
        public TimeSpan CrackTime { get; set; }
        public TimeSpan BandaidTimer { get; set; }
        public TimeSpan BandaidTime { get; set; }

        // Scoring.
        public int ScoringTotalCovered { get; set; }
        public float CurrentScore => -(Wall.MinX - Wall.WALL_INITIAL_X + Wall.WALL_INITIAL_WIDTH) + ScoringTotalCovered*4;

        // Used to restrict player movement for first few frames.
        public int GameFrameCount { get; set; }
        public bool ManualCameraMovement { get; set; }

        public DamGameController(Core.DamGame game, GraphicsController gc, InputController ic, CameraController cc) : base(game, gc, ic)
        {
            _cameraController = cc;

            Wall = new Wall();
            Hero = new Hero(cc, 15, 0);
            Bandaids = new BandaidGraphic[MAX_BANDAIDS];
            Waters = new WaterGraphic[MAX_WATER_PARTICLES];
        }

        public override void DefineStateDefinitions()
        {
            MainMenu = new GameState("mainmenu")
            {
                OnEnterState = () => { },
                OnStateUpdate = MainMenuOnUpdate
            };
            LevelInitialising = new GameState("levelstart")
            {
                OnEnterState = OnLevelStart,
                OnStateUpdate = (state) => SetState(LevelRunning)
            };
            LevelRunning = new GameState("game")
            {
                OnEnterState = () => { },
                OnStateUpdate = GameOnUpdate
            };
            GameOver = new GameState("endgame")
            {
                OnEnterState = () => { },
                OnStateUpdate = GameOnUpdate // Uses the same update loop.
            };
        }

        private void OnLevelStart()
        {
            Console.WriteLine("Initialising level...");

            BackgroundLayers = new BackgroundLayer[7];

            // background 1 ([0]) is the furthest away - the sky
            BackgroundLayers[0] = new BackgroundLayer { MoveSpeedX = 0, ParallaxLevel = 0, Graphic = DamGraphicsController.Background1.GetRandomTexture() };
            BackgroundLayers[1] = new BackgroundLayer { MoveSpeedX = 0.1f, ParallaxLevel = 0.9f, Graphic = DamGraphicsController.Background2.GetRandomTexture() };
            BackgroundLayers[2] = new BackgroundLayer { MoveSpeedX = 0, ParallaxLevel = 0.87f, Graphic = DamGraphicsController.Foreground4.GetRandomTexture() };
            BackgroundLayers[3] = new BackgroundLayer { MoveSpeedX = 0.3f, ParallaxLevel = 0.8f, Graphic = DamGraphicsController.Background3.GetRandomTexture() };
            BackgroundLayers[4] = new BackgroundLayer { MoveSpeedX = 0, ParallaxLevel = 0.7f, Graphic = DamGraphicsController.Foreground3.GetRandomTexture() };
            BackgroundLayers[5] = new BackgroundLayer { MoveSpeedX = 0, ParallaxLevel = 0.6f, Graphic = DamGraphicsController.Foreground2.GetRandomTexture() };
            BackgroundLayers[6] = new BackgroundLayer { MoveSpeedX = 0, ParallaxLevel = 0, Graphic = DamGraphicsController.Foreground.GetRandomTexture() };

            Wall.Reset();

            for (int i = 0; i < MAX_BANDAIDS; i++)
            {
                if (Bandaids[i] == null)
                {
                    Bandaids[i] = new BandaidGraphic(game.GraphicsDevice, new Bandaid(0, 0, true));
                }
                else
                {
                    Bandaids[i].Visible = false;
                }
            }

            for (int i = 0; i < MAX_WATER_PARTICLES; i++)
            {
                if(Waters[i] == null)
                {
                    Waters[i] = new WaterGraphic(game.GraphicsDevice, new Water(0, 0, true));
                }
                else
                {
                    Waters[i].Visible = false;
                }
            }
            
            Hero.Reset(-Core.DamGame.BASE_GAME_RES_WIDTH/2, Core.DamGame.BASE_GAME_RES_HEIGHT, -1, true);
            (graphicsController as DamGraphicsController).Reset();

            _cameraController.MoveToInstant(new Vector2(-Core.DamGame.BASE_GAME_RES_WIDTH, 0));

            GameFrameCount = 0;
            ScoringTotalCovered = 0;
            
            CrackTime = TimeSpan.FromSeconds(2f);
            CrackTimer = TimeSpan.Zero;

            BandaidTime = TimeSpan.FromSeconds(BANDAID_COOLDOWN_IN_SECONDS);
            BandaidTimer = TimeSpan.Zero;
        }

        public void MainMenuOnUpdate(GameTime gameTime)
        {
            var prev = currentMainMenuOption;

            // Update menu selection.
            switch (currentMainMenuOption)
            {
                case MenuOption.NewGame:
                    if (IC.MenuMoveDown.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.Exit;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    if (IC.MoveRight.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.ToggleMute;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    break;
                case MenuOption.Exit:
                    if (IC.MenuMoveUp.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.NewGame;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    if (IC.MoveRight.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.ToggleMute;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    break;
                case MenuOption.ToggleMute:
                    if (IC.MoveLeft.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.Exit;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    if (IC.MenuMoveUp.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.NewGame;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    if (IC.MenuMoveDown.IsJustPressed)
                    {
                        currentMainMenuOption = MenuOption.Exit;
                        //SoundController.Play(VesselSoundController.LowBoop);
                    }

                    break;
            }

            // Initial.
            if (IC.Confirm.IsJustPressed)
            {
                switch (currentMainMenuOption)
                {
                    case MenuOption.NewGame:
                        SetState(LevelInitialising);
                        break;
                    case MenuOption.ToggleMute:
                        SoundController.ToggleMute();
                        SoundController.ToggleMuteMusic();
                        //SoundController.Play(VesselSoundController.HighBoop, false, true);
                        break;
                    case MenuOption.Exit:
                        game.Exit();
                        break;
                }
            }

            if (IC.Cancel.IsJustPressed)
            {
                game.Exit();
            }
        }

        private void SpawnBandaid()
        {
            foreach(var b in Bandaids)
            {
                if (!b.Visible)
                {
                    b.Visible = true;
                    b.Bandaid.Reset(Hero.X + Bandaid.WIDTH / 2, Hero.Y - 20, Hero.Direction == Hero.FaceDirection.Left);

                    break;
                }
            }
        }

        private void SpawnWater(Crack c, int amount)
        {
            int knownVisibleIdx = 0;

            for (int i = 0; i < amount; i++)
            {
                
                for (int j = knownVisibleIdx; j < MAX_WATER_PARTICLES; j++)
                {
                    if (!Waters[j].Visible)
                    {
                        Waters[j].Visible = true;
                        Waters[j].Water.Reset(c.X + (Water.WIDTH / 10) * i, c.Y + (Water.HEIGHT / 10) * i, true); // Always pours leftwards.
                        break;
                    }
                }
                
                // If we reach the end of the loop, we can't spawn a new water. But it doesn't affect gameplay so it's ok.
            }
        }
        
        // Main game logic.
        private void GameOnUpdate(GameTime gameTime)
        {
#if DEBUG_MODE
            int amount = 20;
            if (IC.CameraMoveLeft.IsDown)
            {
                _cameraController.MoveTo(new Vector2(_cameraController.Position.X - amount, _cameraController.Position.Y));
            }
            if (IC.CameraMoveRight.IsDown)
            {
                _cameraController.MoveTo(new Vector2(_cameraController.Position.X + amount, _cameraController.Position.Y));
            }
            if (IC.CameraMoveUp.IsDown)
            {
                _cameraController.MoveTo(new Vector2(_cameraController.Position.X, _cameraController.Position.Y - amount));
            }
            if (IC.CameraMoveDown.IsDown)
            {
                _cameraController.MoveTo(new Vector2(_cameraController.Position.X, _cameraController.Position.Y + amount));
            }

            if (IC.SpawnCrack.IsJustPressed)
            {
                Wall.AddCrack();
            }
            if (IC.ToggleCameraMode.IsJustPressed)
            {
                ManualCameraMovement = !ManualCameraMovement;
            }
            if (IC.ShuffleRenderTargets.IsJustPressed)
            {
                (graphicsController as DamGraphicsController).ShuffleRenderTargets();
            }
#endif

          BandaidTimer += gameTime.ElapsedGameTime;

            if (GameFrameCount > NUM_FRAMES_MOVEMENT_DISABLED)
            {
                Hero.WaitingForLevelToStart = false;
            }

            if (Hero.CanInteract)
            {
                if (IC.MoveLeft.IsDown)
                {
                    Hero.MoveLeft();
                }
                if (IC.MoveRight.IsDown)
                {
                    Hero.MoveRight();
                }

                if (IC.Jump.IsDown)
                {
                    Hero.TryJump();
                }

                if (CurrentState == LevelRunning)
                {
                    if (IC.ThrowBandaid.IsJustPressed)
                    {
                        if (BandaidTimer > BandaidTime)
                        {
                            BandaidTimer = TimeSpan.Zero;
                            SpawnBandaid();
                        }
                    }
                }
            }

            Wall.Update(gameTime);
            Hero.Update(gameTime, this);

            foreach (var b in Bandaids)
            {
                if(b.Visible)
                {
                    b.Bandaid.Update(gameTime, this, _cameraController.CurrentPosition.X);
                }
            }

            foreach (var w in Waters)
            {
                if(w.Visible)
                {
                    w.Water.Update(gameTime, this);
                }
            }

            foreach(var b in Bandaids)
            {
                // Remove bad ones after update
                if (b.Bandaid.isAlmostOnWall && b.Bandaid.isAlmostOnWallFirstFrame)
                {
                    DamSoundController.Play(DamSoundController.WallFix);

                    // TODO combine these loops?
                    // TODO move crack to segment object?
                    // Draw all cracks that are newly covered.
                    for(int i = 0; i < Wall.NUM_SEGMENTS; i++)
                    {
                        var c = Wall.Cracks[i];
                        var segment2 = Wall.Segments[i];
                        var yPos2 = WallSegment.GetSegmentY(i);

                        // We already covered one crack. Here, we check for others.
                        // TODO move the other logic here too?
                        if (c.IsActive && c.IsCovered)
                        {
                            c.IsActive = false;

                            // Now draw the bandaid over the cracks.
                            (graphicsController as DamGraphicsController).AddCrackToWallData(c);

                            ScoringTotalCovered++;
                        }
                        
                        if (yPos2 > b.Bandaid.Y - Bandaid.HEIGHT / 2 && yPos2 < b.Bandaid.Y + Bandaid.HEIGHT / 2)
                        {
                            const int INCREASE_SEGMENT_BY = 15; // originally 10, then 20

                            // Update the segment's X.
                            segment2.X = Math.Min(segment2.X, b.Bandaid.StuckX - INCREASE_SEGMENT_BY);
                        }
                    }
                }

                if (b.Visible && b.Bandaid.isOnWall)
                {
                    // Now draw the bandaid over the cracks.
                    (graphicsController as DamGraphicsController).AddBandaidToWallData(b);

                    // Remove bandaid.
                    b.Visible = false;
                }

                if (b.Bandaid.markForDeletion)
                {
                    // Remove bandaid.
                    b.Visible = false;
                }
            }
            
            for (int i = Waters.Length - 1; i >= 0; i--)
            {
                // Remove bad ones after update
                if (Waters[i].Water.markForDeletion)
                {
                    Waters[i].Visible = false;
                }
            }

            // Update the bg layers.
            foreach (var bl in BackgroundLayers)
            {
                bl.Update(gameTime, _cameraController.CurrentPosition.X);
            }

            // Spawn water.
            foreach(var c in Wall.Cracks)
            {
                if (c.IsActive)
                {
                    var waterAmount = 0;

                    if (c.TimeAlive < Crack.FirstCrackUntil)
                    {
                        if (c.WaterSpawnCounter % 15 == 0)
                        {
                            waterAmount = 1;
                        }
                    }
                    else if (c.TimeAlive < Crack.SecondCrackUntil)
                    {
                        if (c.WaterSpawnCounter % 8 == 0)
                        {
                            waterAmount = 1;
                        }
                    }
                    else if (c.TimeAlive < Crack.ThirdCrackUntil)
                    {
                        if (c.WaterSpawnCounter % 6 == 0)
                        {
                            waterAmount = 2;
                        }
                        else if (c.WaterSpawnCounter % 3 == 0)
                        {
                            waterAmount = 1;
                        }
                    }
                    else if (c.TimeAlive < Crack.FourthCrackUntil)
                    {
                        waterAmount = 3;
                    }
                    else
                    {
                        waterAmount = 5;
                    }
                    
                    SpawnWater(c, waterAmount);
                }
            }

            CrackTimer += gameTime.ElapsedGameTime;

            if (CrackTimer > CrackTime)
            {
                CrackTimer = TimeSpan.Zero;
                CrackTime -= TimeSpan.FromSeconds(0.05f);
                if (CrackTime < TimeSpan.FromSeconds(0.3f))
                {
                    CrackTime = TimeSpan.FromSeconds(0.3f);
                }

                Wall.AddCrack();
            }

            if(!ManualCameraMovement)
            {
                // Update camera position.
                _cameraController.MoveTo(new Vector2(Math.Min(-Core.DamGame.BASE_GAME_RES_WIDTH, -Core.DamGame.BASE_GAME_RES_WIDTH + Wall.MinX + 250), 0));
            }

            if (IC.Cancel.IsJustPressed)
            {
                SetState(MainMenu);
            }

            GameFrameCount++;

            // Lose condition.
            if (Wall.Cracks.Any(c => c.IsFlooded) && CurrentState == LevelRunning)
            {
                SetState(GameOver);
            }

            // Camera position has been updated. Now, see if render targets need to be shuffled.
            (graphicsController as DamGraphicsController).CheckShuffle();
        }
    }
}
