﻿using Dam.Core;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Spl.Core;
using System;
using System.Linq;

namespace Dam.Controllers
{
    public class DamGraphicsController : GraphicsController
    {
        public class WallRenderRect
        {
            public int OffsetX { get; set; }
            public RenderTarget2D RenderTarget { get; set; }
        }

        public const int NUM_WALL_RENDER_TARGETS = 5;
        public const int RENDER_TARGET_WIDTH = 100;
        public const int HEIGHT_BG_HEIGHT = 72;

        public static Graphic Title = new Graphic("title");
        public static Graphic Hero = new Graphic("hero");
        public static Graphic Gameover = new Graphic("gameover");

        public static Graphic Background1 = new Graphic("background-1");
        public static Graphic Background2 = new Graphic("background-2");
        public static Graphic Background3 = new Graphic("background-3");
        public static Graphic Foreground = new Graphic("foreground");
        public static Graphic Foreground2 = new Graphic("foreground-2");
        public static Graphic Foreground3 = new Graphic("foreground-3");
        public static Graphic Foreground4 = new Graphic("foreground-4");

        public static Graphic WallGraphic = new Graphic("dam");
        public static Graphic Bandaid = new Graphic("bandaid");
        public static Graphic Water = new Graphic("water", 2);

        public static Graphic CrackFixed = new Graphic("crack-fixed", 5);
        
        public static Graphic CrackActive1 = new Graphic("crack-active-1");
        public static Graphic CrackActive2 = new Graphic("crack-active-2");
        public static Graphic CrackActive3 = new Graphic("crack-active-3");
        public static Graphic CrackActive4 = new Graphic("crack-active-4");

        public static Font DebugFont = new Font("DebugFont");

        private Texture2D _solidWhite;
        private CameraController _cameraController;

        // Create render targets to draw our wall into. This is our "updatable" Texture2D.
        public WallRenderRect[] RenderTargets { get; set; }

        public SpriteBatch MainBatch { get; set; }

        public DepthStencilState stencilBaseContent = new DepthStencilState
        {
            StencilEnable = true,
            StencilFunction = CompareFunction.Always,
            StencilPass = StencilOperation.Replace,
            ReferenceStencil = 1,
            DepthBufferEnable = false,
        };

        public DepthStencilState stencilDrawOnto = new DepthStencilState
        {
            StencilEnable = true,
            StencilFunction = CompareFunction.LessEqual,
            StencilPass = StencilOperation.Keep,
            ReferenceStencil = 1,
            DepthBufferEnable = false,
        };

        // Check, do we have any render targets out of view of the camera?
        public void CheckShuffle()
        {
            foreach(var r in RenderTargets)
            {
                if (r.OffsetX > _cameraController.CurrentPosition.X + DamGame.BASE_GAME_RES_WIDTH)
                {
                    Console.WriteLine("SHUFFLING A RENDER TARGET. Offset: " + r.OffsetX);
                    ShuffleRenderTarget(r);
                }
            }
        }

        public void ShuffleRenderTarget(WallRenderRect w)
        {
            w.OffsetX -= RENDER_TARGET_WIDTH * NUM_WALL_RENDER_TARGETS;

            // Clear the render target.
            graphicsDevice.SetRenderTarget(w.RenderTarget);
            graphicsDevice.Clear(Color.Transparent);
            graphicsDevice.SetRenderTarget(null);

        }

        // Find the oldest render target and shuffle it.
        public void ShuffleRenderTargets()
        {
            // https://stackoverflow.com/a/3188804/11678918
            var targetToShuffle = RenderTargets.Aggregate((r1, r2) => r1.OffsetX > r2.OffsetX ? r1 : r2);

            targetToShuffle.OffsetX -= RENDER_TARGET_WIDTH * NUM_WALL_RENDER_TARGETS;

            // Clear the render target.
            graphicsDevice.SetRenderTarget(targetToShuffle.RenderTarget);
            graphicsDevice.Clear(Color.Transparent);
            graphicsDevice.SetRenderTarget(null);

        }

        static DamGraphicsController() { } // Added to ensure static Graphic properties are initialised.
        public DamGraphicsController(Game game, GraphicsDevice gd, CameraController cc) : base(game, gd)
        {
            MainBatch = new SpriteBatch(game.GraphicsDevice);
            
            _solidWhite = new Texture2D(graphicsDevice, 1, 1);
            _solidWhite.SetData(new Color[] { Color.White });

            _cameraController = cc;
        }

        public void AfterLoad()
        {
            RenderTargets = new WallRenderRect[NUM_WALL_RENDER_TARGETS];

            for (int i = 0; i < NUM_WALL_RENDER_TARGETS; i++)
            {
                RenderTargets[i] = new WallRenderRect
                {
                    OffsetX = 0,
                    RenderTarget = new RenderTarget2D(graphicsDevice,
                        RENDER_TARGET_WIDTH,
                        graphicsDevice.PresentationParameters.BackBufferHeight,
                        false,
                        graphicsDevice.PresentationParameters.BackBufferFormat,
                        graphicsDevice.PresentationParameters.DepthStencilFormat,
                        graphicsDevice.PresentationParameters.MultiSampleCount,
                        RenderTargetUsage.PreserveContents)
                };
            }
        }

        public void Reset()
        {
            for (int i = 0; i < NUM_WALL_RENDER_TARGETS; i++)
            {
                RenderTargets[i].OffsetX = -RENDER_TARGET_WIDTH * (i + 1);
            }

            DrawInitialWallData();
        }

        public override void SetGameStateDrawFunctions()
        {
            DamGameController.MainMenu.Draw = DrawMenu;
            DamGameController.LevelRunning.Draw = DrawGame;
            DamGameController.GameOver.Draw = DrawGameOverMenu;
        }

        public void DrawMenu()
        {
            MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend);
            {
                // The initial wall graphic.
                MainBatch.Draw(Title.GetRandomTexture(), new Vector2(0, 0), Color.White);
            }
            MainBatch.End();
        }

        public void DrawGameOverMenu()
        {
            DrawGame();

            MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend);
            {
                // The initial wall graphic.
                MainBatch.Draw(Gameover.GetRandomTexture(), new Vector2(0, 0), Color.White);
            }
            MainBatch.End();
        }

        public void DrawInitialWallData()
        {
            var cameraMatrix = _cameraController.Transform;
            var vgc = gameController as DamGameController;
            
            foreach(var t in RenderTargets)
            {
                graphicsDevice.SetRenderTarget(t.RenderTarget);
                graphicsDevice.Clear(Color.Transparent);

                // Draw the wall.
                // Wall graphics. Cracks will be overlayed onto this content.
                MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend,
                    depthStencilState: stencilBaseContent);//, effect: a);
                {
                    // The initial wall graphic.
                    MainBatch.Draw(
                        WallGraphic.GetRandomTexture(),
                        new Vector2(-t.OffsetX + Wall.WALL_INITIAL_X - Wall.WALL_INITIAL_WIDTH, Wall.WALL_TOP_Y),
                        Color.White);
                }
                MainBatch.End();
            }
            
            graphicsDevice.SetRenderTarget(null);
        }

        public void AddBandaidToWallData(BandaidGraphic b)
        {
            var cameraMatrix = _cameraController.Transform;
            
            var vgc = gameController as DamGameController;

            foreach(var target in RenderTargets)
            {
                // Only render a new bandaid onto the texture when asked...
                // Activate it.
                graphicsDevice.SetRenderTarget(target.RenderTarget);
                
                MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend,
                    depthStencilState: stencilBaseContent);
                {
                    // Draw the bandaid.
                    MainBatch.Draw(b.Graphic,
                           new Vector2(-target.OffsetX + b.Bandaid.realX, b.Bandaid.realY),
                           sourceRectangle: null, color: Color.White,
                           rotation: b.Bandaid.rotation, origin: new Vector2(Dam.Bandaid.WIDTH / 2, Dam.Bandaid.HEIGHT / 2), scale: 1,
                           effects: SpriteEffects.None, layerDepth: 0);
                }
                MainBatch.End();
            }
            
            // Switch back to normal render target.
            graphicsDevice.SetRenderTarget(null);
        }

        public void AddCrackToWallData(Crack c)
        {
            var cameraMatrix = _cameraController.Transform;
            var vgc = gameController as DamGameController;

            foreach (var target in RenderTargets)
            {
                // Activate it.
                graphicsDevice.SetRenderTarget(target.RenderTarget);

                // Draw an inactive crack.
                MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend,
                    depthStencilState: stencilDrawOnto);
                {
                    MainBatch.Draw(CrackFixed.GetRandomTexture(), new Vector2(-target.OffsetX + c.X, c.Y + Crack.GRAPHIC_Y_OFFSET), Color.White);
                }
                MainBatch.End();
            }

            // Switch back to normal render target.
            graphicsDevice.SetRenderTarget(null);
        }

        public void DrawGame()
        {
            var cameraMatrix = _cameraController.Transform;
            
            var vgc = gameController as DamGameController;

            // Render the wall graphic based on current camera transform.
            var m = Matrix.CreateOrthographicOffCenter(
                graphicsDevice.PresentationParameters.BackBufferWidth + _cameraController.CurrentPosition.X,
                graphicsDevice.PresentationParameters.BackBufferWidth*2 + _cameraController.CurrentPosition.X,
                graphicsDevice.PresentationParameters.BackBufferHeight + _cameraController.CurrentPosition.Y,
                0 + _cameraController.CurrentPosition.Y,
                0, 1
            );

            using (var a = new AlphaTestEffect(graphicsDevice) { Projection = m })
            {
                // Draw background
                MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend, transformMatrix: cameraMatrix);
                {
                    // Background layers.
                    foreach (var bl in vgc.BackgroundLayers)
                    {
                        MainBatch.Draw(bl.Graphic, new Vector2(bl.X - DamGame.BASE_GAME_RES_WIDTH * 2, 0), Color.White); // Fixed.
                    }

                    const int WATER_HEIGHT_BELOW_WALL = 20;

                    // Water.
                    MainBatch.Draw(_solidWhite,
                        new Rectangle(Wall.WALL_INITIAL_X, Wall.WALL_TOP_Y + WATER_HEIGHT_BELOW_WALL, 1000, Wall.WALL_HEIGHT - WATER_HEIGHT_BELOW_WALL),
                        new Color(21, 48, 169));

                    // Current water drops.
                    foreach (var w in vgc.Waters)
                    {
                        if(w.Visible)
                        {
                            MainBatch.Draw(w.Graphic,
                                new Vector2(w.Water.X, w.Water.Y), sourceRectangle: null,
                                color: Color.White, rotation: w.Water.rotation, origin: new Vector2(Dam.Water.WIDTH / 2, Dam.Water.HEIGHT / 2), scale: 1,
                                effects: SpriteEffects.None, layerDepth: 0);
                        }
                    }
                }
                MainBatch.End();

                // Clear the stencil data. Fixes smearing effect when camera scrolls and cracks get revealed.
                graphicsDevice.Clear(ClearOptions.Stencil, Color.Transparent, 0, 0);

                // Draw the wall as stencil data.
                // Cracks will be overlayed onto this content.
                // FIXME use offset value as well...
                foreach(var target in RenderTargets)
                {
                    MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend,
                        depthStencilState: stencilBaseContent, effect: a);
                    {
                        MainBatch.Draw(target.RenderTarget, new Vector2(DamGame.BASE_GAME_RES_WIDTH + target.OffsetX, 0), Color.White);
                    }
                    MainBatch.End();
                }

                // Draw the active cracks.
                MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend,
                    depthStencilState: stencilDrawOnto, effect: a);
                {
                    // Draw the ACTIVE cracks. (TODO inactive cracks should be baked into the texture).
                    foreach (var c in vgc.Wall.Cracks)
                    {
                        if(c.IsActive)
                        {
                            if (c.TimeAlive < Crack.FirstCrackUntil)
                            {
                                MainBatch.Draw(CrackActive1.GetRandomTexture(), new Vector2(800 + c.X, c.Y + Crack.GRAPHIC_Y_OFFSET), Color.White);
                            }
                            else if (c.TimeAlive < Crack.SecondCrackUntil)
                            {
                                MainBatch.Draw(CrackActive2.GetRandomTexture(), new Vector2(800 + c.X, c.Y + Crack.GRAPHIC_Y_OFFSET), Color.White);
                            }
                            else if (c.TimeAlive < Crack.ThirdCrackUntil)
                            {
                                MainBatch.Draw(CrackActive3.GetRandomTexture(), new Vector2(800 + c.X, c.Y + Crack.GRAPHIC_Y_OFFSET), Color.White);
                            }
                            else
                            {
                                MainBatch.Draw(CrackActive4.GetRandomTexture(), new Vector2(800 + c.X, c.Y + Crack.GRAPHIC_Y_OFFSET), Color.White);
                            }
                        }
                    }
                }
                MainBatch.End();
            }

            MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend, transformMatrix: cameraMatrix);
            {
                // The player.
                MainBatch.Draw(Hero.GetRandomTexture(), new Vector2(vgc.Hero.X, vgc.Hero.Y - Dam.Hero.HERO_HEIGHT), Color.White);

                // Draw the bandaids.
                foreach (var b in vgc.Bandaids)
                {
                    if (b.Visible)
                    {
                        MainBatch.Draw(b.Graphic,
                            new Vector2(b.Bandaid.X, b.Bandaid.Y), sourceRectangle: null,
                            color: Color.White, rotation: b.Bandaid.rotation, origin: new Vector2(Dam.Bandaid.WIDTH / 2, Dam.Bandaid.HEIGHT / 2), scale: 1,
                            effects: SpriteEffects.None, layerDepth: 0);
                    }
                }

#if DEBUG_MODE
                for(int i = 0; i < NUM_WALL_RENDER_TARGETS; i++)
                {
                    MainBatch.Draw(_solidWhite, new Rectangle(RenderTargets[i].OffsetX, 10 * (i+1), RENDER_TARGET_WIDTH, 10), Color.Purple);
                }
#endif
            }
            MainBatch.End();

            MainBatch.Begin(samplerState: SamplerState.LinearClamp, blendState: BlendState.AlphaBlend);
            {
                MainBatch.Draw(_solidWhite, new Rectangle(0, 600 - HEIGHT_BG_HEIGHT, 220, HEIGHT_BG_HEIGHT), Color.White);

                MainBatch.DrawString(DebugFont.SpriteFont, "Patches: " + vgc.ScoringTotalCovered, new Vector2(10, 600 - HEIGHT_BG_HEIGHT), Color.Black);
                MainBatch.DrawString(DebugFont.SpriteFont, $"Distance: " + (-(vgc.Wall.MinX - Wall.WALL_INITIAL_X + Wall.WALL_INITIAL_WIDTH) / 150f) + "m", new Vector2(10, 600 - HEIGHT_BG_HEIGHT + 22), Color.Black);
                MainBatch.DrawString(DebugFont.SpriteFont, "Score: " + vgc.CurrentScore, new Vector2(10, 600 - HEIGHT_BG_HEIGHT + 44), Color.Black);
            }
            MainBatch.End();
        }
    }
}
